--[[
    [Code Doodle] Snowclock
    @version 1.0, 2017-04-06
    @author TheOddByte
--]]


--# Gets the dimensions of the window
local width, height = love.graphics.getDimensions()

--# The characters used for the clock
local characters = require( "characters" )


--# Misc Variables
--# Most of them control the speed
--# of different things
local gridlines   = false;
local spawnTimer  = 0.8
local spawnRate   = { 0.05, 0.1 }
local fallTimer   = 0.04
local normal      = 0.04
local fast        = 0.02
local smoothTimer = 0
local wind        = 0


--# The color IDs
--# 0 = background
--# 1 = snow
local colors = {
    ["0"] = { 35, 35, 75 };
    ["1"] = { 255, 255, 255 };
    ["2"] = { 250, 35, 35 };
}


--# The grid
local oldTime
local cache = {}
local size = 4
local w, h = width/size, height/size
local grid = {}
for x = 1, w do
    grid[x] = {}
    for y = 1, h do
        grid[x][y] = "0"
    end
end


local function getTime()
    return os.date( "%H:%M" )
end


local function updateTime()
    if cache then
        local time = os.date( "%H:%M" )
        for i = 1, #text do
            local char = text:sub( i, i )
            for _y, line in ipairs( characters[ char ] ) do
                for _x = 1, #line do
                    local c      = line:sub( _x, _x )
                    local gX, gY = (x+_x-1), (y+_y-1)
                    grid[ gX ][ gY ] = c == "#" and "2" or "0"
                end
            end
            x = x + #characters[ char ][1] + 1
        end
    end
end


local function getDimensions( text )
    local len = 0
    for i = 1, #text do
        len = len + #characters[ text:sub( i, i ) ][1] + 1
    end
    return len, #characters["1"]
end


local function pr( text, x, y )
    for i = 1, #text do
        local char = text:sub( i, i )
        for _y, line in ipairs( characters[ char ] ) do
            for _x = 1, #line do
                local c      = line:sub( _x, _x )
                local gX, gY = (x+_x-1), (y+_y-1)
                if c == "#" then
                    grid[ gX ][ gY ] = "2"
                else
                    if grid[ gX ][ gY ] == "2" then
                        grid[ gX ][ gY ] = "0"
                    end
                end
            end
        end
        x = x + #characters[ char ][1] + 1
    end
end



--# Moves/Switches place for two points
--# in the grid
local function move( x1, y1, x2, y2 )
    if grid[x1][y1] ~= "2" and grid[x2][y2] ~= "2" then
        local a = grid[x1][y1]
        local b = grid[x2][y2]

        grid[x1][y1] = b
        grid[x2][y2] = a
    end
end


--# Spawns a new snowflake at the top
--# of the grid
local function spawn()
    grid[ math.random( 1, w ) ][ 1 ] = "1"
end


--[[
    Find the snowflakes that are still
    falling down.

    @return table points
--]]
local function findFalling()
    local points = {}
    for x = 1, w do
        for y = 1, h - 1 do
            if grid[x][y] == "1" and grid[x][ y + 1 ] == "0" then
                table.insert( points, { x, y } )
            end
        end
    end
    return points
end



--[[
    Find the snowflakes that are still
    falling down.

    @return table points
--]]
local function findFallen()
    local points = {}
    for x = 1, w do
        for y = 1, h - 1 do
            if grid[x][y] == "1" and grid[x][ y + 1 ] == "1" then
                table.insert( points, { x, y } )
            end
        end
    end
    return points
end



local function clearGrid()
    w    = width/size
    h    = height/size
    grid = {}
    for x = 1, w do
        grid[x] = {}
        for y = 1, h do
            grid[x][y] = "0"
        end
    end
end



local function setSize( s )
    size = s
    clearGrid()
end



local function transform( x, y )
    return math.ceil(( x - 1 ) * size), math.ceil( ( y - 1 ) * size )
end



local function t( x, y )
    for _x = 1, w do
        for _y = 1, h do
            local rX, rY = transform( _x, _y )
            if x >= rX and x <= rX + size and y >= rY and y <= rY + size then
                return _x, _y
            end
        end
    end
end


local function getPoint( x, y )
    return grid[x][y]
end



function love.load()
    love.window.setMode( 1024, 384 )
    width, height = love.graphics.getDimensions()
    clearGrid()
    --love.window.setMode( w*size, (h*size) )
    love.window.setTitle( "Code Doodle - Snowclock" )
end



function love.update( dt )

    --# Control the wind
    wind = love.keyboard.isDown( "left" ) and -1 or love.keyboard.isDown( "right" ) and 1 or 0

    if not oldTime or getTime() ~= oldTime then
        local tw, th = getDimensions( getTime() )
        pr( getTime(), math.ceil( (w-tw)/2 ), math.ceil( (h-th) / 2 ) )
        oldTime = getTime()
    end

    spawnTimer = spawnTimer - dt
    if spawnTimer <= 0 then
        spawnTimer = math.random( spawnRate[1], spawnRate[2] )
        spawn()
    end

    fallTimer = fallTimer - dt
    if fallTimer <= 0 then
        fallTimer = ( wind == 0 and not love.keyboard.isDown( "down" ) ) and normal or fast
        local points = findFalling()
        if #points > 0 then
            for _, point in ipairs( points ) do
                local valid = {}
                local x, y  = point[1], point[2]

                if x - 1 >= 1 and grid[ x - 1][ y + 1] == "0" then table.insert( valid, -1 ) end
                if x + 1 <= w and grid[ x + 1][ y + 1] == "0" then table.insert( valid, 1 ) end
                table.insert( valid, 0 )

                local dX = valid[ math.random( #valid ) ] + wind
                local nX = x + dX < 1 and 1 or x + dX > w and w or x + dX
                if grid[ nX ][ y + 1 ] == "0" then
                    move( x, y, nX, y + 1 )
                end
            end
        end
    end

    --# Smoothen
    local points = findFallen()
    if smoothTimer <= 0 then
        smoothTimer = 0
        if #points > 0 then
            smoothTimer = 0.145
        end
    else
        smoothTimer = smoothTimer - dt
        if smoothTimer <= 0 then
            smoothTimer = 0
            for _, point in ipairs( points ) do
                local valid = {}
                local x, y  = point[1], point[2]
                if x - 1 >= 1 then
                    if grid[ x - 1 ][ y ] ~= "1" then
                        if grid[ x - 1 ][ y + 1 ] ~= "1" then
                            table.insert( valid, -1 )
                        end
                    end
                end
                if x + 1 <= w then
                    if grid[ x + 1 ][ y ] ~= "1" then
                        if grid[ x + 1 ][ y + 1 ] ~= "1" then
                            table.insert( valid, 1 )
                        end
                    end
                end

                if #valid > 0 then
                    local dX = valid[ math.random( #valid ) ]
                    move( x, y, x + dX, y + 1 )
                end
            end
        end
    end
end


function love.mousepressed(x, y, button, isTouch)
    local gX, gY = t( x, y )
    grid[gX][gY] = "2"
end

function love.keypressed(key, scancode, isrepeat)
    if key == "r" then
        clearGrid()

    elseif key == "g" then
        gridlines = not gridlines

    elseif key == "+" then
        setSize( size*2 )
        spawnRate[1] = spawnRate[1]*2
        spawnRate[2] = spawnRate[2]*2
        normal       = normal*2
        fast         = fast*2

    elseif key == "-" then
        setSize( size/2 )
        spawnRate[1] = spawnRate[1]/2
        spawnRate[2] = spawnRate[2]/2
        normal       = normal/2
        fast         = fast/2

    elseif key == "t" then
        print( os.date("%H:%M:%S") )

    end
end



function love.draw()
    for x = 1, w do
        for y = 1, h do
            local color  = colors[ grid[x][y] ]
            local rX, rY = transform( x, y )
            love.graphics.setColor( color[1], color[2], color[3], 255 )
            love.graphics.rectangle( "fill", rX, rY, size, size )
            if gridlines then
                love.graphics.setColor( 15, 15, 15, 200 )
                love.graphics.rectangle( "line", rX, rY, size, size )
            end
        end
    end
end
